﻿using Microsoft.AspNetCore.Hosting;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using NPOI.HSSF.UserModel;
using NPOI.SS.UserModel;
using NPOI.SS.Util;
using NPOI.XSSF.UserModel;
using Performance.DtoModels;
using Performance.DtoModels.AppSettings;
using Performance.EntityModels;
using Performance.Infrastructure;
using Performance.Repository;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace Performance.Services
{
    /// <summary>
    /// 数据抽取服务
    /// </summary>
    public class ExtractService : IAutoInjection
    {
        private readonly ILogger<ExtractService> logger;
        private readonly IHostingEnvironment environment;
        private readonly IEmailService emailService;
        private readonly PerSheetService perSheetService;
        private readonly PerHeaderService perHeaderService;
        private readonly PerforPersheetRepository perforPersheetRepository;
        private readonly PerforImheaderRepository perforImheaderRepository;
        private readonly PerforImemployeeRepository perforImemployeeRepository;
        private readonly PerforExtractRepository perforExtractRepository;
        private readonly PerforPerfirstRepository perforPerfirstRepository;
        private readonly PerforPerallotRepository perforPerallotRepository;
        private readonly PerforHospitalconfigRepository perforHospitalconfigRepository;

        public ExtractService(ILogger<ExtractService> logger,
            IHostingEnvironment environment,
            IEmailService emailService,
            PerSheetService perSheetService,
            PerHeaderService perHeaderService,
            PerforPersheetRepository perforPersheetRepository,
            PerforImheaderRepository perforImheaderRepository,
            PerforImemployeeRepository perforImemployeeRepository,
            PerforExtractRepository perforExtractRepository,
            PerforPerfirstRepository perforPerfirstRepository,
            PerforPerallotRepository perforPerallotRepository,
            PerforHospitalconfigRepository perforHospitalconfigRepository)
        {
            this.logger = logger;
            this.environment = environment;
            this.emailService = emailService;
            this.perSheetService = perSheetService;
            this.perHeaderService = perHeaderService;
            this.perforPersheetRepository = perforPersheetRepository;
            this.perforImheaderRepository = perforImheaderRepository;
            this.perforImemployeeRepository = perforImemployeeRepository;
            this.perforExtractRepository = perforExtractRepository;
            this.perforPerfirstRepository = perforPerfirstRepository;
            this.perforPerallotRepository = perforPerallotRepository;
            this.perforHospitalconfigRepository = perforHospitalconfigRepository;
        }

        public void ExtractData(int allotId, string mail, sys_hospital hospital)
        {
            List<PerSheet> sheetList = new List<PerSheet>();

            try
            {
                var allotList = perforPerallotRepository.GetEntities(t => t.HospitalId == hospital.ID);
                var configList = perforHospitalconfigRepository.GetEntities(t => t.HospitalId == hospital.ID);
                var firstList = perforPerfirstRepository.GetEntities(t => t.HospitalId == hospital.ID);
                var scriptList = perforExtractRepository.GetEntities(t => t.HospitalId == hospital.ID);

                if (configList == null || !configList.Any())
                {
                    LogHelper.Error("暂不支持自动提取绩效数据", "提取绩效数据");
                    throw new PerformanceException($"暂不支持自动提取绩效数据");
                }

                if (scriptList == null || !scriptList.Any())
                {
                    LogHelper.Error($"暂无该医院的配置信息和执行sql，无法进行数据提取", "提取绩效数据");
                    throw new PerformanceException($"暂无该医院的配置信息和执行sql，无法进行数据提取");
                }

                var hospitalConfig = configList.First();
                string originalPath = "";
                //判断是否是首次
                if ((allotList == null || !allotList.Any(t => t.States > 0)) && (firstList != null || !firstList.Any()))
                {
                    var first = firstList.OrderByDescending(t => t.CreateDate).FirstOrDefault();
                    //首次 从excel中获取人员信息，SHEET页信息，列头信息
                    sheetList = GetFileData(first.Path);
                    originalPath = first.Path;
                }
                else
                {
                    //非首次 从数据库中获取人员信息，SHEET页信息，列头信息
                    allotList = allotList.Where(t => t.Path != null && t.Path != "").ToList();
                    if (allotList != null && allotList.Count > 0)
                    {
                        var allot = allotList.OrderByDescending(t => t.Year).ThenByDescending(t => t.Month).First();
                        sheetList = GetRepositoryData(allot.ID);
                        originalPath = allot.Path;
                    }
                    else
                    {
                        LogHelper.Error($"历史绩效未上传文件", "提取绩效数据");
                        throw new PerformanceException($"历史绩效未上传文件");
                    }
                }
                var dpath = Path.Combine(environment.ContentRootPath, "Files", $"{hospital.ID}", "autoextract");
                FileHelper.CreateDirectory(dpath);
                string path = Path.Combine(dpath, $"绩效数据{DateTime.Now.ToString("yyyyMMddHHmmssfff")}.xlsx");
                //根据SHEET页信息，列头信息，创建EXCEL文件
                if (WriteExcel(path, originalPath, sheetList, hospitalConfig, hospital.ID, out string filepath))
                {
                    var allot = perforPerallotRepository.GetEntity(t => t.ID == allotId);
                    allot.ExtractPath = filepath;
                    if (!string.IsNullOrEmpty(filepath))
                        perforPerallotRepository.Update(allot);

                    SendEmail(mail, filepath, $"{hospital.HosName}HIS数据提取成功", $"{hospital.HosName}在{DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss")}成功提取。");
                }
            }
            catch (PerformanceException ex)
            {
                logger.LogError(ex.ToString());
                LogHelper.Error(ex.ToString(), "提取绩效数据", "异常");
                SendEmail(mail, "", $"{hospital.HosName}HIS数据提取失败", $"{hospital.HosName}提取数据失败，{ex.Message}！");
            }
            catch (Exception ex)
            {
                logger.LogError(ex.ToString());
                LogHelper.Error(ex.ToString(), "提取绩效数据", "异常");
                SendEmail(mail, "", $"{hospital.HosName}HIS数据提取失败", $"{hospital.HosName}提取数据过程中出现异常情况，我们将尽快解决问题。给您带来的不便我们深感歉意！");
            }
        }

        /// <summary>
        /// 发送邮件
        /// </summary>
        /// <param name="path"></param>
        /// <param name="subject"></param>
        /// <param name="body"></param>
        private void SendEmail(string mail, string path, string subject, string body)
        {
            var message = new EmailMessage
            {
                To = new List<string> { mail },
                DisplayName = "溯直健康",
                Subject = subject,
                Body = body
            };
            if (!string.IsNullOrEmpty(path))
                message.Attachments = new List<string> { path };
            emailService.Send(message);
        }

        /// <summary>
        /// 向EXCEL中写入数据
        /// 人员信息由EXCEL中提供，或从上次绩效中提供
        /// SHEET页在SQL提取中出现时，执行SQL脚本获得结果向EXCEL中填充
        /// 向EXCEL中填充数据，填充时值与列头必须匹配
        /// </summary>
        /// <param name="newpath"></param>
        /// <param name="originalPath"></param>
        /// <param name="sheetList"></param>
        /// <param name="hospitalConfig"></param>
        /// <param name="hospitalId"></param>
        private bool WriteExcel(string newpath, string originalPath, List<PerSheet> sheetList, sys_hospitalconfig hospitalConfig, int hospitalId, out string filepath)
        {
            if (string.IsNullOrEmpty(originalPath))
            {
                LogHelper.Error($"{originalPath}文件路径无效", "提取绩效数据");
                throw new PerformanceException($"{originalPath}文件路径无效");
            }

            var scriptList = perforExtractRepository.GetEntities(t => t.HospitalId == hospitalId);

            //根据SHEET页信息，列头信息，创建EXCEL文件
            IWorkbook workbook = null;
            try
            {
                workbook = new XSSFWorkbook(originalPath);
                foreach (var sheet in sheetList)
                {
                    var importSheet = workbook.GetSheet(sheet.SheetName);
                    if (importSheet == null) continue;
                    if (sheet.PerHeader == null) continue;

                    var maxHeaderRowNumber = sheet.PerHeader.Max(t => t.PointRow);
                    sheet.PerHeader?.ForEach(t =>
                    {
                        if (t.IsHasChildren)
                        {
                            var maxnum = t.Children.Max(c => c.PointRow);
                            maxHeaderRowNumber = maxHeaderRowNumber > maxnum ? maxHeaderRowNumber : maxnum;
                        }
                    });
                    if (sheet.SheetType == SheetType.Workload)
                        maxHeaderRowNumber += 1;
                    //清空数据行
                    for (int i = maxHeaderRowNumber + 1; i < importSheet.LastRowNum + 1; i++)
                    {
                        var importRow = importSheet.GetRow(i);
                        if (importRow != null)
                            importSheet.RemoveRow(importRow);
                    }

                    //填充人员信息
                    if (SheetType.Employee == sheet.SheetType && sheet.PerData != null && sheet.PerData.Any())
                    {
                        var dataList = sheet.PerData.ConvertAll(new Converter<IPerData, PerDataEmployee>(t => (PerDataEmployee)t));
                        for (int i = 0; i < dataList.Count; i++)
                        {
                            var importRow = importSheet.CreateRow(maxHeaderRowNumber + i + 1);
                            Dictionary<string, Func<PerDataEmployee, object>> keyValues = new Dictionary<string, Func<PerDataEmployee, object>>
                        {
                            { "核算单元", (t) => t.AccountingUnit },
                            { "医生姓名", (t) => t.DoctorName },
                            { "职称", (t) => t.JobTitle },
                            { "绩效基数核算参考对象", (t) => t.FitPeople },
                            { "核算单元分类", (t) => t.AccountType },
                            { "岗位系数", (t) => t.PostCoefficient },
                            { "参加工作时间", (t) => t.WorkTime },
                            { "考核得分率", (t) => t.ScoreAverageRate },
                            { "出勤率", (t) => t.Attendance },
                            { "核算单元医生数", (t) => t.PeopleNumber },
                            { "工作量绩效", (t) => t.Workload },
                            { "其他绩效", (t) => t.OthePerfor },
                            { "医院奖罚", (t) => t.Punishment },
                            { "调节系数", (t) => t.Adjust },
                            { "发放系数", (t) => t.Grant },
                        };
                            foreach (var item in keyValues.Keys)
                            {
                                var headInfo = sheet.PerHeader.FirstOrDefault(t => t.CellValue == item);
                                if (headInfo != null)
                                {
                                    var value = (keyValues[item].Invoke(dataList[i]) ?? "").ToString();
                                    importRow.CreateCell(headInfo.PointCell).SetCellValue(Verify(value));
                                }
                            }
                        }
                    }
                    //SHEET页在SQL提取中出现时，执行SQL脚本获得结果向EXCEL中填充
                    else if (scriptList.Any(t => sheet.SheetName.Contains(t.SheetName)))
                    {
                        var script = scriptList.First(t => sheet.SheetName.Contains(t.SheetName));
                        if (!string.IsNullOrEmpty(script.ExecuteScript))
                        {
                            var children = new List<PerHeader>();
                            foreach (var item in sheet.PerHeader?.Select(t => t.Children))
                            {
                                if (item != null)
                                    children = children.Union(item).ToList();
                            }

                            var connection = ConnectionBuilder.Create(DatabaseType.SqlServer, hospitalConfig.DbSource, hospitalConfig.DbName, hospitalConfig.DbUser, hospitalConfig.DbPassword);
                            var dataList = perforExtractRepository.ExecuteScript(connection, script.ExecuteScript, null);
                            //创建数据行
                            foreach (var pointRow in dataList.Select(t => t.RowNumber).Distinct().OrderBy(t => t))
                            {
                                var importRow = importSheet.CreateRow(maxHeaderRowNumber + pointRow + 1);
                                //写入单元格
                                foreach (var data in dataList.Where(t => t.RowNumber == pointRow))
                                {
                                    var headInfo = sheet.PerHeader.FirstOrDefault(t => t.CellValue == data.ColumnName)
                                        ?? children?.FirstOrDefault(t => t.CellValue == data.ColumnName);
                                    if (headInfo != null)
                                    {
                                        importRow.CreateCell(headInfo.PointCell).SetCellValue(Verify(data.Value.ToString()));
                                    }
                                }
                            }
                        }
                    }
                }
                using (FileStream file = new FileStream(newpath, FileMode.Create))
                    workbook.Write(file);
                filepath = newpath;
                return true;
            }
            catch (Exception ex)
            {
                throw ex;
            }
            finally
            {
                workbook.Close();
            }
        }

        /// <summary>
        /// 从数据库中获取sheet及列头
        /// </summary>
        /// <param name="allotId"></param>
        /// <returns></returns>
        private List<PerSheet> GetRepositoryData(int allotId)
        {
            List<PerSheet> sheetList = new List<PerSheet>();
            //获取最近一次绩效
            var perSheetList = perforPersheetRepository.GetEntities(t => t.AllotID == allotId);
            if (perSheetList == null || perSheetList.Count == 0)
                return sheetList;


            foreach (var sheet in perSheetList)
            {
                PerSheet perSheet = new PerSheet()
                {
                    SheetName = sheet.SheetName,
                    SheetType = (SheetType)sheet.SheetType,
                    ModuleName = EnumHelper.GetDescription((SheetType)sheet.SheetType),
                };
                var perHeadList = perforImheaderRepository.GetEntities(t => t.SheetID == sheet.ID);
                perSheet.PerHeader = AutoMapper.Mapper.Map<List<PerHeader>>(perHeadList);
                if (SheetType.Employee == (SheetType)sheet.SheetType)
                {
                    perSheet.PerHeader = GetHeader((SheetType)sheet.SheetType);
                    var employeeList = perforImemployeeRepository.GetEntities(t => t.AllotID == sheet.AllotID);
                    var perEmployeeList = AutoMapper.Mapper.Map<List<PerDataEmployee>>(employeeList);
                    perSheet.PerData = perEmployeeList?.ConvertAll(new Converter<PerDataEmployee, IPerData>(m => m));
                }
                else if (SheetType.SpecialUnit == (SheetType)sheet.SheetType)
                {
                    perSheet.PerHeader = GetHeader((SheetType)sheet.SheetType);
                }
                else if (SheetType.AccountBasic == (SheetType)sheet.SheetType)
                {
                    perSheet.PerHeader = GetHeader((SheetType)sheet.SheetType);
                }
                sheetList.Add(perSheet);
            }
            return sheetList;
        }



        /// <summary>
        /// 获取sheet及列头
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        private List<PerSheet> GetFileData(string path)
        {
            if (!FileHelper.IsExistFile(path))
                throw new PerformanceException($"{path}文件不存在");
            List<PerSheet> sheetList = new List<PerSheet>();
            FileStream fs = null;
            IWorkbook workbook = null;
            try
            {
                fs = new FileStream(path, FileMode.Open, FileAccess.Read);
                var version = FileHelper.GetExtension(path) == ".xlsx" ? ExcelVersion.xlsx : ExcelVersion.xls;
                workbook = (version == ExcelVersion.xlsx) ? (IWorkbook)(new XSSFWorkbook(fs)) : (IWorkbook)(new HSSFWorkbook(fs));
                for (int i = 0; i < workbook.NumberOfSheets; i++)
                {
                    var sheet = workbook.GetSheetAt(i);
                    var sheetType = perSheetService.GetSheetType(sheet.SheetName);

                    if (SheetType.Unidentifiable != sheetType)
                    {
                        var sheetRead = PerSheetDataFactory.GetDataRead(sheetType);
                        var perHeader = perHeaderService.GetPerHeader(sheet, sheetRead.Point);

                        PerSheet perSheet = new PerSheet()
                        {
                            SheetName = sheet.SheetName,
                            SheetType = sheetType,
                            ModuleName = EnumHelper.GetDescription(sheetType),
                            PerHeader = perHeader
                        };
                        if (SheetType.Employee == sheetType)
                        {
                            var headerReverse = perHeaderService.GetPerHeaderReverse(perHeader);
                            perSheet.PerData = sheetRead.ReadData(sheet, headerReverse);
                        }
                        sheetList.Add(perSheet);
                    }
                }
            }
            catch (Exception ex)
            {
                throw ex;
            }
            finally
            {
                workbook.Close();
                fs.Close();
            }

            return sheetList;
        }

        /// <summary>
        /// 校验数据格式，并转换
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public dynamic Verify(string obj)
        {
            try
            {
                //判断值是否为double类型
                if (!string.IsNullOrEmpty(obj) && Regex.Match(obj.Trim(), @"([1-9]\d*\.?\d*)|(0\.\d*[1-9]?\d*)|0").ToString() == obj.Trim())
                    return ConvertHelper.To<double>(obj);
                //判断值是否为日期格式
                else if (!string.IsNullOrEmpty(obj) && Regex.Match(obj.Trim(), @"(19|20)\d{2}(-|/)[01]?\d(-|/)[0123]?\d( [012]?\d\:\d{2}\:\d{2})?").ToString() == obj.Trim())
                    return ConvertHelper.To<DateTime>(obj).ToString("yyyy/MM/dd");
                else
                    return obj;
            }
            catch
            {
                return obj;
            }
        }

        #region 获取列头
        /// <summary>
        /// 获取列头
        /// </summary>
        /// <param name="sheetType">sheet页类型</param>
        /// <returns></returns>
        public List<PerHeader> GetHeader(SheetType sheetType)
        {
            Dictionary<string, List<int>> keyValues = new Dictionary<string, List<int>>();
            if (SheetType.Employee == sheetType)
            {
                keyValues = new Dictionary<string, List<int>>
                    {
                        { "核算单元分类",new List<int> { 1, 1, 0, 1 } },
                        { "核算单元", new List<int> { 1, 1, 1, 1 } },
                        { "绩效基数核算参考对象", new List<int> { 1, 1, 2, 1 } },
                        { "医生姓名", new List<int> { 1, 1, 3, 1 } },
                        { "职称", new List<int> { 1, 1, 4, 1 } },
                        { "岗位系数", new List<int> { 1, 1, 5, 1 } },
                        { "参加工作时间", new List<int> { 1, 1, 6, 1 } },
                        { "考核得分率", new List<int> { 1, 1, 7, 1 } },
                        { "出勤率", new List<int> { 1, 1, 8, 1 } },
                        { "核算单元医生数", new List<int> { 1, 1, 9, 1 } },
                        { "工作量绩效", new List<int> { 1, 1, 10, 1 } },
                        { "其他绩效", new List<int> { 1, 1, 11, 1 } },
                        { "医院奖罚", new List<int> { 1, 1, 12, 1 } },
                        { "调节系数", new List<int> { 1, 1, 13, 1 } },
                        { "发放系数", new List<int> { 1, 1, 14, 1 } },
                    };
            }
            else if (SheetType.AccountBasic == sheetType)
            {
                keyValues = new Dictionary<string, List<int>>
                    {
                        { "核算单元",new List<int> { 1, 2, 0, 1 } },
                        { "科室", new List<int> { 1, 2, 1, 1 } },
                        { "医生组", new List<int> { 8, 1, 2, 1 } },
                        { "护理组", new List<int> { 8, 1, 3, 1 } },
                        { "核算单元医生数量", new List<int> { 1, 1, 2, 2 } },
                        { "基础系数", new List<int> { 1, 1, 3, 2 } },
                        { "倾斜系数", new List<int> { 1, 1, 4, 2 } },
                        { "其他绩效1", new List<int> { 1, 1, 5, 2 } },
                        { "考核得分率", new List<int> { 1, 1, 6, 2 } },
                        { "医院奖罚", new List<int> { 1, 1, 7, 2 } },
                        { "其他绩效2", new List<int> { 1, 1, 8, 2 } },
                        { "调节系数", new List<int> { 1, 1, 9, 2 } },
                        { "核算单元护士数量", new List<int> { 1, 1, 10, 2 } },
                        { "基础系数?", new List<int> { 1, 1, 11, 2 } },
                        { "倾斜系数?", new List<int> { 1, 1, 12, 2 } },
                        { "其他绩效1?", new List<int> { 1, 1, 13, 2 } },
                        { "考核得分率?", new List<int> { 1, 1, 14, 2 } },
                        { "医院奖罚?", new List<int> { 1, 1, 15, 2 } },
                        { "其他绩效2?", new List<int> { 1, 1, 16, 2 } },
                        { "调节系数?", new List<int> { 1, 1, 17, 2 } },
                    };
            }
            else if (SheetType.SpecialUnit == sheetType)
            {
                keyValues = new Dictionary<string, List<int>>
                    {
                        { "科室",new List<int> { 1, 1, 0, 1 } },
                        { "人数", new List<int> { 1, 1, 1, 1 } },
                        { "量化指标", new List<int> { 1, 1, 2, 1 } },
                        { "数量", new List<int> { 1, 1, 3, 1 } },
                        { "量化指标绩效分值", new List<int> { 1, 1, 4, 1 } },
                        { "考核得分率", new List<int> { 1, 1, 5, 1 } },
                        { "医院奖罚", new List<int> { 1, 1, 6, 1 } },
                        { "其他绩效", new List<int> { 1, 1, 7, 1 } },
                        { "调节系数", new List<int> { 1, 1, 8, 1 } },
                    };
            }

            List<PerHeader> header = new List<PerHeader>();
            foreach (var item in keyValues.Keys)
            {
                header.Add(new PerHeader
                {
                    CellValue = item.Replace("?", ""),
                    IsTotal = 0,
                    MergeCell = keyValues[item][0],
                    MergeRow = keyValues[item][1],
                    PointCell = keyValues[item][2],
                    PointRow = keyValues[item][3]
                });
            }
            return header;
        }
        #endregion
    }
}
